/******************************************************************************
 * GViD Source Code Editor
 * Copyright (C) 2002  Alan Oursland
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Alan Oursland
 * alan@oursland.net
 *****************************************************************************/
package gvid.ui.editor;

import java.awt.*;
import java.awt.datatransfer.*;
import gvid.Globals;
import gvid.ui.FindSettings;
import gvid.ui.editor.*;
import gvid.ui.editor.undo.*;
import gvid.ui.editor.selection.*;
import gvid.undo.*;
import gvid.document.*;
import gvid.ui.*;
import gvid.undo.UndoChain;

public class EditorEditCommandsImpl implements gvid.command.EditCommands
{
	SourceCodeEditor m_editor = null;
	public EditorEditCommandsImpl( SourceCodeEditor editor ) {
		m_editor = editor;
	}
	
	public int getCommandGroup() {
		return EDITGROUP;
	}
	
	public UndoCommand charLeft() {
		// get the current view position
		// find the document position
		Point d = m_editor.posToDoc(m_editor.getCaretPosition());
		// adjust the document position
		if( d.x > 0 ) {
			d.x--;
		} else if( d.y>0 ) {
			d.y--;
			d.x = m_editor.getLineLength(d.y);
		}
		// convert back to view position
		Point newP = m_editor.docToPos(d);
		// build undo command
		UndoCommand rc = m_editor.setCaretPosition(newP, true);
		rc.setName("charLeft");
		return rc;
	}

	public UndoCommand charLeftExtend() {
		// find the current document position
		Point d = m_editor.posToDoc(m_editor.getCaretPosition());
		// adjust the document position
		if( d.x > 0 ) {
			d.x--;
		} else if( d.y>0 ) {
			d.y--;
			d.x = m_editor.getLineLength(d.y);
		}
		
		// convert back to view position
		Point newP = m_editor.docToPos(d);
		
		// set the new selection
		UndoChain rc = new UndoChain("");
		rc.addRedo(m_editor.setSelectionMode(SourceCodeSelection.BOUNDED));
		rc.addRedo(m_editor.setSelectionPosition(newP));
		rc.setName("charLeftExtend");
		return rc;
	}

	public UndoCommand charRight() {
		// get the current view position
		// find the document position
		Point d = m_editor.posToDoc(m_editor.getCaretPosition());
		// adjust the document position
		if( d.x < m_editor.getLineLength(d.y) ) {
			d.x++;
		} else if( d.y<m_editor.getLineCount() ) {
			d.y++;
			d.x = 0;
		}
		// convert back to view position
		Point newP = m_editor.docToPos(d);
		// build undo command
		UndoCommand rc = m_editor.setCaretPosition(newP, true);
		rc.setName("charRight");
		return rc;
	}

	public UndoCommand charRightExtend() {
		// find the current document position
		Point d = m_editor.posToDoc(m_editor.getCaretPosition());
		// adjust the document position
		if( d.x < m_editor.getLineLength(d.y) ) {
			d.x++;
		} else if( d.y<m_editor.getLineCount() ) {
			d.y++;
			d.x = 0;
		}
		
		// convert back to view position
		Point newP = m_editor.docToPos(d);

		// set the new selection
		UndoChain rc = new UndoChain("");
		rc.addRedo(m_editor.setSelectionMode(SourceCodeSelection.BOUNDED));
		rc.addRedo(m_editor.setSelectionPosition(newP));
		rc.setName("charRightExtend");
		
		return rc;
	}

	public UndoCommand copy() {
		if( m_editor.getSelection() != null ) {
			SourceCodeSelection s = m_editor.getSelection().copy();
			Clipboard gcb = Toolkit.getDefaultToolkit().getSystemClipboard();
			gcb.setContents(s, s);
		}
		// nothing to undo
		return null;
	}

	public UndoCommand cut() {
		UndoCommand rc = null;
		if( m_editor.getSelection() != null ) {
			copy();
			rc = delete();
			rc.setName("cut");
		}
		return rc;
	}

	public UndoCommand delete() {
		UndoChain rc = new UndoChain("delete");
		Point p = m_editor.getCaretPosition();
		Point d = m_editor.posToDoc(p);
		p = m_editor.docToPos(d);
		if( m_editor.getSelection() == null ) {
			if( p.y < (m_editor.getLineCount()) || p.x < m_editor.getLineLength(p.y) ) {
				rc.addRedo(m_editor.setCaretPosition(p, true));	
					// in case the true caret position is to the right of the line
					// and we combine lines. This will set the caret position to the
					// real place where we edited.
				rc.addRedo(m_editor.getSourceModel().deleteChar(d));
			}
		} else {
			rc.addRedo(m_editor.getSelection().delete());
			rc.addRedo(m_editor.clearSelection());
		}
		return rc;
	}

	public UndoCommand deleteBack() {
		UndoChain rc = new UndoChain("deleteBack");
		if( m_editor.getSelection() == null ) {
			if( !m_editor.getCaretPosition().equals(new Point(0,0)) ) {
				rc.addRedo(charLeft());
				rc.addRedo(delete());		
			}
		} else {
			rc.addRedo(delete());
		}
		return rc;
	}

	public UndoCommand documentEnd() {
		int lastY = m_editor.getLineCount();
		int lastX = m_editor.getLineLength(lastY);
		Point newP = m_editor.docToPos(new Point(lastX, lastY));
		UndoCommand rc = m_editor.setCaretPosition(newP, true);
		rc.setName("documentEnd");
		return rc;
	}

	public UndoCommand documentEndExtend() {
		int lastY = m_editor.getLineCount();
		int lastX = m_editor.getLineLength(lastY);
		Point newP = m_editor.docToPos(new Point(lastX, lastY));
		// set the new selection
		UndoChain rc = new UndoChain("");
		rc.addRedo(m_editor.setSelectionMode(SourceCodeSelection.BOUNDED));
		rc.addRedo(m_editor.setSelectionPosition(newP));
		rc.setName("documentEndExtend");

		return rc;
	}

	public UndoCommand documentStart() {
		UndoCommand rc = m_editor.setCaretPosition(new Point(0,0), true);
		rc.setName("documentStart");
		return rc;
	}

	public UndoCommand documentStartExtend() {
		Point newP = new Point(0, 0);
		// set the new selection
		UndoChain rc = new UndoChain("");
		rc.addRedo(m_editor.setSelectionMode(SourceCodeSelection.BOUNDED));
		rc.addRedo(m_editor.setSelectionPosition(newP));
		rc.setName("documentStartExtend");
		
		return rc;
	}

	public UndoCommand editToggleOvertype() {
		if( EditCaret.isOverwrite() ) {
			EditCaret.setInsert();
		} else {
			EditCaret.setOverwrite();
		}
		return null;
	}

	public UndoCommand endBrief() {
		UndoCommand rc = null;
		Point d = m_editor.posToDoc(m_editor.getCaretPosition());
		if( !m_editor.getSourceModel().isEOL(d) ) {
			rc = lineEnd();			
		} else if( m_editor.getLastVisibleRow() != d.y ) {
			rc = windowEnd();
		} else {
			rc = documentEnd();
		}
		return rc;
	}

	public boolean isAvailable( int cmd ) {
		boolean rc = false;
		
		// TODO: Implement isAvailable
		rc = true;
		
		return rc;
	}
	
	public void executeCommand(int cmd) {
		UndoCommand undo = null;
		switch(cmd) {
			case REDO: redo(); break;
			case REDOCHANGES: redoChanges(); break;
			case UNDO: undo(); break;
			case UNDOCHANGES: undoChanges(); break;
			case CHARLEFT: undo = charLeft(); break;
			case WORDLEFT: undo = wordLeft(); break;
			case SENTENCELEFT: undo = sentenceLeft(); break;
			case HOME: undo = home(); break;
			case WINDOWLEFTEDGE: undo = windowLeftEdge(); break;
			case GOTOINDENTATION: undo = goToIndentation(); break;
			case CHARRIGHT: undo = charRight(); break;
			case WORDRIGHT: undo = wordRight(); break;
			case SENTENCERIGHT: undo = sentenceRight(); break;
			case LINEEND: undo = lineEnd(); break;
			case WINDOWRIGHTEDGE: undo = windowRightEdge(); break;
			case LINEUP: undo = lineUp(); break;
			case PARAUP: undo = paraUp(); break;
			case PAGEUP: undo = pageUp(); break;
			case DOCUMENTSTART: undo = documentStart(); break;
			case WINDOWSTART: undo = windowStart(); break;
			case HOMEBRIEF: undo = homeBrief(); break;
			case LINEDOWN: undo = lineDown(); break;
			case PARADOWN: undo = paraDown(); break;
			case PAGEDOWN: undo = pageDown(); break;
			case DOCUMENTEND: undo = documentEnd(); break;
			case WINDOWEND: undo = windowEnd(); break;
			case ENDBRIEF: undo = endBrief(); break;
			case WINDOWSCROLLUP: undo = windowScrollUp(); break;
			case WINDOWSCROLLDOWN: undo = windowScrollDown(); break;
			case WINDOWSCROLLLEFT: undo = windowScrollLeft(); break;
			case WINDOWSCROLLRIGHT: undo = windowScrollRight(); break;
			case WINDOWSCROLLTOTOP: undo = windowScrollToTop(); break;
			case WINDOWSCROLLTOBOTTOM: undo = windowScrollToBottom(); break;
			case WINDOWSCROLLTOCENTER: undo = windowScrollToCenter(); break;
			case EDITTOGGLEOVERTYPE: undo = editToggleOvertype(); break;
			case TOGGLEVIEWWHITESPACE: undo = toggleViewWhitespace(); break;
			case CHARLEFTEXTEND: undo = charLeftExtend(); break;
			case CHARRIGHTEXTEND: undo = charRightExtend(); break;
			case DOCUMENTENDEXTEND: undo = documentEndExtend(); break;
			case DOCUMENTSTARTEXTEND: undo = documentStartExtend(); break;
			case HOMEEXTEND: undo = homeExtend(); break;
			case LINEDOWNEXTEND: undo = lineDownExtend(); break;
			case LINEENDEXTEND: undo = lineEndExtend(); break;
			case LINEUPEXTEND: undo = lineUpExtend(); break;
			case PAGEDOWNEXTEND: undo = pageDownExtend(); break;
			case PAGEUPEXTEND: undo = pageUpExtend(); break;
			case SELECTALL: undo = selectAll(); break;
//			case SELECTLINE: undo = selectLine(); break;
			case SELECTPARA: undo = selectPara(); break;
			case SELECTSWAPANCHOR: undo = selectSwapAnchor(); break;
			case WORDLEFTEXTEND: undo = wordLeftExtend(); break;
			case WORDRIGHTEXTEND: undo = wordRightExtend(); break;
			case SPLITLINE: undo = splitLine(); break;
			case DELETE: undo = delete(); break;
			case DELETEBACK: undo = deleteBack(); break;
			case COPY: undo = copy(); break;
			case CUT: undo = cut(); break;
			case PASTE: undo = paste(); break;
			case WORDDELETETOEND: undo = wordDeleteToEnd(); break;	
			case WORDDELETETOBEGINNING: undo = wordDeleteToBeginning(); break;
			case SELECTIONLOWERCASE: undo = selectionLowercase(); break;
			case SELECTIONUPPERCASE: undo = selectionUppercase(); break;
			case SELECTIONTOGGLECASE: undo = selectionTogglecase(); break;
			case SELECTIONINDENT: undo = selectionIndent(); break;
			case SELECTIONUNINDENT: undo = selectionUnindent(); break;
			case INDENTTOPREV: undo = indentToPrev(); break;
			case CLEARSELECTION: undo = clearSelection(); break;
			case FIND: undo = find(); break;
			case FINDNEXT: undo = findNext(); break;
			case FINDPREV: undo = findPrev(); break;
			case FINDSELECTIONNEXT: undo = findSelectionNext(); break;
			case FINDSELECTIONPREV: undo = findSelectionPrev(); break;
		}
		if( undo != null ) {
			m_editor.getUndoManager().addUndo(undo);
		}
	}

	public UndoCommand goToIndentation() {
		Point newP = m_editor.posToDoc(m_editor.getCaretPosition());
		newP.x = m_editor.getSourceModel().lineFirstWord(newP.y);
		newP = m_editor.docToPos(newP);
		UndoCommand rc = m_editor.setCaretPosition(newP, true);
		rc.setName("goToIndentation");
		return rc;
	}

	public UndoCommand home() {
		Point newP = m_editor.posToDoc(m_editor.getCaretPosition());
		int homePos = m_editor.getSourceModel().lineFirstWord(newP.y);
		if( newP.x == homePos ) {
			newP.x = 0;
		} else {
			newP.x = homePos;
		}
		newP = m_editor.docToPos(newP);
		UndoCommand rc = m_editor.setCaretPosition(newP, true);
		rc.setName("home");
		return rc;
	}

	public UndoCommand homeBrief() {
		UndoCommand rc = null;
		Point d = m_editor.posToDoc(m_editor.getCaretPosition());
		if( d.x != 0 ) {
			rc = home();			
		} else if( m_editor.getFirstVisibleRow() != d.y ) {
			rc = windowStart();
		} else {
			rc = documentStart();
		}
		return rc;
	}

	public UndoCommand homeExtend() {
		Point newP = m_editor.posToDoc(m_editor.getCaretPosition());
		int homePos = m_editor.getSourceModel().lineFirstWord(newP.y);
		if( newP.x == homePos ) {
			newP.x = 0;
		} else {
			newP.x = homePos;
		}
		newP = m_editor.docToPos(newP);
		// set the new selection
		UndoChain rc = new UndoChain("");
		rc.addRedo(m_editor.setSelectionMode(SourceCodeSelection.BOUNDED));
		rc.addRedo(m_editor.setSelectionPosition(newP));
		rc.setName("homeExtend");
		return rc;
	}

	public UndoCommand lineDown() {
		Point p = m_editor.getCaretPosition();
		p.y++;
		UndoCommand rc = m_editor.setCaretPosition(p, true);
		rc.setName("lineDown");
		return rc;
	}

	public UndoCommand lineDownExtend() {
 		Point p = m_editor.getCaretPosition();
 		p.y++;
		// set the new selection
		UndoChain rc = new UndoChain("");
		rc.addRedo(m_editor.setSelectionMode(SourceCodeSelection.BOUNDED));
		rc.addRedo(m_editor.setSelectionPosition(p));
		rc.setName("lineDownExtend");
		return rc;
   }

	public UndoCommand lineEnd() {
		// get the current view position
		Point p = m_editor.getCaretPosition();
		// find the document position
		Point d = m_editor.posToDoc(p);
		// set document position to end of line
		d.x = m_editor.getLineLength(d.y);
		// convert back to view position
		Point newP = m_editor.docToPos(d);
		UndoCommand rc = m_editor.setCaretPosition(newP, true);
		rc.setName("lineEnd");
		return rc;		
	}

	public UndoCommand lineEndExtend() {
		// get the current view position
		Point p = m_editor.getCaretPosition();
		// find the document position
		Point d = m_editor.posToDoc(p);
		// set document position to end of line
		d.x = m_editor.getLineLength(d.y);
		// convert back to view position
		Point newP = m_editor.docToPos(d);
		// set the new selection
		UndoChain rc = new UndoChain("");
		rc.addRedo(m_editor.setSelectionMode(SourceCodeSelection.BOUNDED));
		rc.addRedo(m_editor.setSelectionPosition(newP));
		rc.setName("lineEndExtend");
		return rc;
	}

	public UndoCommand lineUp() {
		Point p = m_editor.getCaretPosition();
		p.y--;
		UndoCommand rc = m_editor.setCaretPosition(p, true);
		rc.setName("lineUp");
		return rc;
	}

	public UndoCommand lineUpExtend() {
		Point p = m_editor.getCaretPosition();
		Point newP = new Point(p.x, p.y-1);
		// set the new selection
		UndoChain rc = new UndoChain("");
		rc.addRedo(m_editor.setSelectionMode(SourceCodeSelection.BOUNDED));
		rc.addRedo(m_editor.setSelectionPosition(newP));
		rc.setName("lineUpExtend");

		return rc;
	}

	public UndoCommand pageDown() {
		Point p = m_editor.getCaretPosition();
		int scrollValue = m_editor.getFirstVisibleRow();
		p.y += m_editor.getVisibleRowCount();
		UndoCommand rc = m_editor.setCaretPosition(p, true);
		rc.setName("pageDown");
		m_editor.setFirstVisibleRow(scrollValue+m_editor.getVisibleRowCount());
		return rc;
	}

	public UndoCommand pageDownExtend() {
		Point p = m_editor.getCaretPosition();
		Point newP = new Point(p.x, p.y+m_editor.getVisibleRowCount());
		int scrollValue = m_editor.getFirstVisibleRow();
		// set the new selection
		UndoChain rc = new UndoChain("");
		rc.addRedo(m_editor.setSelectionMode(SourceCodeSelection.BOUNDED));
		rc.addRedo(m_editor.setSelectionPosition(newP));
		rc.setName("pageDownExtend");
		m_editor.setFirstVisibleRow(scrollValue+m_editor.getVisibleRowCount());
		return rc;
   }

	public UndoCommand pageUp() {
		Point p = m_editor.getCaretPosition();
		int scrollValue = m_editor.getFirstVisibleRow();
		p.y -= m_editor.getVisibleRowCount();
		UndoCommand rc = m_editor.setCaretPosition(p, true);
		rc.setName("pageUp");
		m_editor.setFirstVisibleRow(scrollValue-m_editor.getVisibleRowCount());
		return rc;
	}

	public UndoCommand pageUpExtend() {
		Point p = m_editor.getCaretPosition();
		Point newP = new Point(p.x, p.y-m_editor.getVisibleRowCount());
		int scrollValue = m_editor.getFirstVisibleRow();
		// set the new selection
		UndoChain rc = new UndoChain("");
		rc.addRedo(m_editor.setSelectionMode(SourceCodeSelection.BOUNDED));
		rc.addRedo(m_editor.setSelectionPosition(newP));
		rc.setName("pageUpExtend");
		m_editor.setFirstVisibleRow(scrollValue-m_editor.getVisibleRowCount());
		return rc;
	}

	public UndoCommand paraDown() {
		Point p = m_editor.getCaretPosition();
		Point newP = m_editor.posToDoc(p);
		// Move past remaining blank lines
		while( newP.y<m_editor.getLineCount() && m_editor.getLineLength(newP.y) == 0 ) {
			newP.y++;
		} 
		// Find next blank line
		while( newP.y<m_editor.getLineCount() && m_editor.getLineLength(newP.y) > 0 ) {
			newP.y++;
		} 
		newP = m_editor.docToPos(newP);
		newP.x = 0;
		UndoCommand rc = m_editor.setCaretPosition(newP, true);
		rc.setName("paraDown");
		return rc;		
	}

	public UndoCommand paraUp() {
		Point p = m_editor.getCaretPosition();
		Point newP = m_editor.posToDoc(p);
		// Move past remaining blank lines
		newP.y--;
		while( newP.y>m_editor.getFirstLine() && m_editor.getLineLength(newP.y) == 0 ) {
			newP.y--;
		} 
		// Find next blank line
		while( newP.y>=m_editor.getFirstLine() && m_editor.getLineLength(newP.y) > 0 ) {
			newP.y--;
		} 
		
		// we actually want the last line with text
		newP.y++;
		
		newP = m_editor.docToPos(newP);
		newP.x = 0;
		UndoCommand rc = m_editor.setCaretPosition(newP, true);
		rc.setName("paraUp");
		return rc;
	}

	public UndoCommand paste() {
		if( m_editor.getSelection() != null ) {
			delete();
		}
		Clipboard gcb = Toolkit.getDefaultToolkit().getSystemClipboard();
		Transferable t = gcb.getContents(this);
		
		DataFlavor flavor = m_editor.getPreferredDataFlavor(t.getTransferDataFlavors());
		Object data = m_editor.getPreferredTransferableObject(t, flavor);
		UndoCommand rc = m_editor.insertTransferableObject(data, flavor);
		rc.setName("paste");
		return rc;
	}

	public void redo() {
		m_editor.getUndoManager().redo(m_editor);
	}

	public void redoChanges() {
		m_editor.getUndoManager().redoChanges(m_editor);
   }

	public UndoCommand selectAll() {
		documentEnd();
		documentStartExtend();
		UndoCommand rc = documentStartExtend();
		rc.setName("selectAll");
		return rc;
   }

	public UndoCommand selectPara() {
		paraUp();
		Point p1 = m_editor.getCaretPosition();
		paraDown();
		Point p2 = m_editor.getCaretPosition();
		m_editor.setCaretPosition(p1, true);
		UndoChain rc = new UndoChain("");
		rc.addRedo(m_editor.setSelectionMode(SourceCodeSelection.BOUNDED));
		rc.addRedo(m_editor.setSelectionPosition(p2));
		rc.setName("selectPara");
		return rc;
   }

	public UndoCommand selectSwapAnchor() {
		SelectionCommand rc = new SelectionCommand( "selectSwapAnchor" );
		rc.initSelection(m_editor);

		if( m_editor.getSelection() != null ) {
			m_editor.getSelection().swapAnchor();
			m_editor.setCaretPosition(m_editor.getSelection().getSelectionEnd(), false);
		}

		rc.finalSelection(m_editor);
		return rc;
   }

	public UndoCommand selectionLowercase() {
		UndoCommand rc = null;
		if( m_editor.getSelection() != null ) {
			rc = m_editor.getSelection().toLowercase();
		}
 		return rc;
   }

	public UndoCommand selectionTogglecase() {
		UndoCommand rc = null;
		if( m_editor.getSelection() != null ) {
			rc = m_editor.getSelection().toggleCase();
		}
 		return rc;
   }

	public UndoCommand selectionUppercase() {
		UndoCommand rc = null;
		if( m_editor.getSelection() != null ) {
			rc = m_editor.getSelection().toUppercase();
		}
 		return rc;
   }

	public UndoCommand sentenceLeft() {
		UndoCommand rc = null;
		// DEFER: implement sentence left
		return rc;
   }

	public UndoCommand sentenceRight() {
		UndoCommand rc = null;
		// DEFER: implement sentence right
		return rc;
	}

	public UndoCommand splitLine() {
		UndoChain rc = new UndoChain("splitLine");
		if( m_editor.isInsert() ) {
	 		if( m_editor.getSelection() != null ) {
				rc.addRedo(delete());
			}
			Point p = m_editor.getCaretPosition();
			Point d = m_editor.posToDoc(p);
			rc.addRedo(m_editor.getSourceModel().split(d));
	   	} else if( m_editor.isOverwrite() ) {
 			Point p = m_editor.getCaretPosition();
			Point d = m_editor.posToDoc(p);
			if( d.y < m_editor.getLineCount() ) {
				// set to beginning of next line
				rc.addRedo(m_editor.setCaretPosition(new Point(0, d.y+1), true));
			} else {
				// set to end of line
				rc.addRedo( m_editor.setCaretPosition( m_editor.docToPos( new Point(m_editor.getLineLength(d.y), d.y) ), true) );
				// add a new line
				rc.addRedo(m_editor.getSourceModel().split(d));
			}
	   	}
		rc.addRedo(indentToPrev());
		return rc;
	}

	public UndoCommand toggleViewWhitespace() {
		if( gvid.Globals.getTabDisplay() == gvid.Globals.INVISIBLETAB ) {
			gvid.Globals.setTabDisplay(true);
		} else {
			gvid.Globals.setTabDisplay(false);
		}
 		return null; // no undo
   }

	public void undo() {
		m_editor.getUndoManager().undo(m_editor);
	}

	public void undoChanges() {
		m_editor.getUndoManager().undoChanges(m_editor);
	}

	public UndoCommand windowEnd() {
		UndoCommand rc = null;
		Point p = m_editor.getCaretPosition();
		if( m_editor.getFirstVisibleRow()+m_editor.getVisibleRowCount() != p.y ) {
			p.y = m_editor.getFirstVisibleRow()+m_editor.getVisibleRowCount();
			rc = m_editor.setCaretPosition(p, true);
			rc.setName("windowEnd");
		}
		return rc;
	}

	public UndoCommand windowLeftEdge() {
		UndoCommand rc = null;
		Point p = m_editor.getCaretPosition();
		if( m_editor.getFirstVisibleColumn() != p.x ) {
			p.x = m_editor.getFirstVisibleColumn();
			rc = m_editor.setCaretPosition(p, true);
			rc.setName("windowLeftEdge");
		}
 		return rc;
   }

	public UndoCommand windowRightEdge() {
		UndoCommand rc = null;
		Point p = m_editor.getCaretPosition();
		if( m_editor.getVisibleColumnCount() != p.x ) {
			p.x = m_editor.getLastVisibleColumn();
			rc = m_editor.setCaretPosition(p, true);
			rc.setName("windowRightEdge");
		}
 		return rc;
   }

	public UndoCommand windowScrollDown() {
		m_editor.setFirstVisibleRow(m_editor.getFirstVisibleRow()-1);
 		return null; // no undo for scrolling
   }

	public UndoCommand windowScrollLeft() {
		m_editor.setFirstVisibleColumn(m_editor.getFirstVisibleColumn()-1);
 		return null; // no undo for scrolling
   }

	public UndoCommand windowScrollRight() {
		m_editor.setFirstVisibleColumn(m_editor.getFirstVisibleColumn()+1);
 		return null; // no undo for scrolling
   }

	public UndoCommand windowScrollToBottom() {
		m_editor.setFirstVisibleRow(m_editor.getCaretPosition().y-m_editor.getVisibleRowCount());
 		return null; // no undo for scrolling
   }

	public UndoCommand windowScrollToCenter() {
		m_editor.setFirstVisibleRow(m_editor.getCaretPosition().y-(m_editor.getVisibleRowCount()/2));
		return null; // no undo for scrolling
	}

	public UndoCommand windowScrollToTop() {
		m_editor.setFirstVisibleRow(m_editor.getCaretPosition().y);
		return null; // no undo for scrolling
	}

	public UndoCommand windowScrollUp() {
		m_editor.setFirstVisibleRow(m_editor.getFirstVisibleRow()+1);
		return null; // no undo for scrolling
	}

	public UndoCommand windowStart() {
		UndoCommand rc = null;
		Point p = m_editor.getCaretPosition();
		if( m_editor.getFirstVisibleRow() != p.y ) {
			p.y = m_editor.getFirstVisibleRow();
			rc = m_editor.setCaretPosition(p, true);
			rc.setName("windowStart");
		}
		return rc;
	}

	public UndoCommand wordDeleteToBeginning() {
		Point p = m_editor.getCaretPosition();
		Point d = m_editor.posToDoc(p);
		int startOfWord = m_editor.getSourceModel().lineWordStart(d.y, d.x);
		Point start = m_editor.docToPos(d);
		Point end = m_editor.docToPos(new Point(startOfWord, d.y));
		UndoCommand rc = m_editor.deleteRange(start,end);
		rc.setName("wordDeleteToBeginning");
		m_editor.repaint();
 		return rc;
   }

	public UndoCommand wordDeleteToEnd() {
		Point p = m_editor.getCaretPosition();
		Point d = m_editor.posToDoc(p);
		int endOfWord = m_editor.getSourceModel().lineWordEnd(d.y, d.x);
		Point start = m_editor.docToPos(d);
		Point end = m_editor.docToPos(new Point(endOfWord, d.y));
		UndoCommand rc = m_editor.deleteRange(start,end);
		rc.setName("wordDeleteToEnd");
		m_editor.repaint();
 		return rc;
   }

	public UndoCommand wordLeft() {
		Point p = m_editor.getCaretPosition();
		Point d = m_editor.posToDoc(p);
		d = m_editor.getSourceModel().prevWord(d);
		d = m_editor.docToPos(d);
		UndoCommand rc = m_editor.setCaretPosition(d, true);
		rc.setName("wordLeft");
		return rc;
	}

	public UndoCommand wordLeftExtend() {
		Point p = m_editor.getCaretPosition();
		Point d = m_editor.posToDoc(p);
		d = m_editor.getSourceModel().prevWord(d);
		d = m_editor.docToPos(d);
		// set the new selection
		UndoChain rc = new UndoChain("");
		rc.addRedo(m_editor.setSelectionMode(SourceCodeSelection.BOUNDED));
		rc.addRedo(m_editor.setSelectionPosition(d));
		rc.setName("wordLeftExtend");
		return rc;
   }

	public UndoCommand wordRight() {
		Point d = m_editor.posToDoc(m_editor.getCaretPosition());
		d = m_editor.getSourceModel().nextWord(d);
		d = m_editor.docToPos(d);
		UndoCommand rc = m_editor.setCaretPosition(d, true);
		rc.setName("wordRight");
		return rc;
	}

	public UndoCommand wordRightExtend() {
		Point p = m_editor.getCaretPosition();
		Point d = m_editor.posToDoc(p);
		d = m_editor.getSourceModel().nextWord(d);
		d = m_editor.docToPos(d);
		// set the new selection
		UndoChain rc = new UndoChain("");
		rc.addRedo(m_editor.setSelectionMode(SourceCodeSelection.BOUNDED));
		rc.addRedo(m_editor.setSelectionPosition(d));
		rc.setName("wordRightExtend");
		return rc;
   }

	public UndoCommand selectionIndent() {
		UndoChain rc = new UndoChain("selectionIndent");
		if( m_editor.getSelection() == null ) {
			// inset the tab character		
			if( m_editor.isInsert() ) {
				rc.addRedo(m_editor.insertChar('\t'));
				rc.setName("insertTab");
			} else if( m_editor.isOverwrite() ) {
				// Move to the next tab stop
				Point p = m_editor.getCaretPosition();
				Point d = m_editor.posToDoc(p);
				p = m_editor.docToPos(d);
				p.x += m_editor.getTabWidth() - p.x%m_editor.getTabWidth();
				rc.addRedo(m_editor.setCaretPosition(p, true));
				rc.setName("nextTab");
			}
		} else {
			// find the current selection positions
			Point selFront = m_editor.posToDoc(m_editor.getSelection().getSelectionStart());
			Point selBack = m_editor.posToDoc(m_editor.getSelection().getSelectionEnd());
			if( isBefore(selBack, selFront) ) {
				Point temp = selFront;
				selFront = selBack;
				selBack = temp;
			}
			SelectionCommand selCmd = new SelectionCommand("selectionIndent");
			selCmd.initSelection(m_editor);
			selCmd.finalSelection(m_editor);
			rc.addRedo(selCmd);
			// insert the tab character at the front of each line
			for(int i=selFront.y; i<=selBack.y; i++ ) {
				if( i!=selBack.y || selBack.x!=0 ) {
					String backId = m_editor.getUnusedReferenceId();
					m_editor.addReferencePoint(selBack, backId, null);
					rc.addRedo(m_editor.getSourceModel().insertChar('\t', new Point(0,i)));
					selBack.setLocation(m_editor.getSourceModel().getReferencePoint(backId));
					m_editor.removeReferencePoint(backId, null);
				}
			}
			// update the selection
			selFront = m_editor.docToPos(selFront);
			if( selBack.x == 0 ) {
				selBack.y--;
			}
			selBack = m_editor.docToPos(new Point(m_editor.getLineLength(selBack.y), selBack.y));
			m_editor.setCaretPosition(new Point(0, selFront.y), true);
			rc.addRedo(m_editor.setSelectionMode(SourceCodeSelection.BOUNDED));
			rc.addRedo(m_editor.setSelectionPosition(selBack));
		}
		return rc;
	}
	
	public UndoCommand selectionUnindent() {
		UndoChain rc = new UndoChain("selectionUnindent");
		if( m_editor.getSelection() == null ) {
			// Move to the previous tab stop
			Point p = m_editor.getCaretPosition();
			Point d = m_editor.posToDoc(p);
			p = m_editor.docToPos(d);
			int tabDistance = p.x%m_editor.getTabWidth();
			if( tabDistance == 0 ) {
				tabDistance = m_editor.getTabWidth();
			}
			p.x -= tabDistance;
			rc.addRedo(m_editor.setCaretPosition(p, true));
			rc.setName("previousTab");
		} else {
			// get the current selection positions
			Point selFront = m_editor.posToDoc(m_editor.getSelection().getSelectionStart());
			Point selBack = m_editor.posToDoc(m_editor.getSelection().getSelectionEnd());
			if( isBefore(selBack, selFront) ) {
				Point temp = selFront;
				selFront = selBack;
				selBack = temp;
			}
			SelectionCommand selCmd = new SelectionCommand("selectionIndent");
			selCmd.initSelection(m_editor);
			selCmd.finalSelection(m_editor);
			rc.addRedo(selCmd);
			// Unindent each line
			for(int i=selFront.y; i<=selBack.y; i++ ) {
				if( i!=selBack.y || selBack.x!=0 ) {
					for( int j=0; j < m_editor.getTabWidth(); j++ ) {
						int c = m_editor.getSourceModel().charAt(new Point(0,i));
						if( c!= -1 && Character.isWhitespace((char)c) ) {
							if( c == '\t' ) {
								j += m_editor.getTabWidth();
							}
							rc.addRedo(m_editor.getSourceModel().deleteChar(new Point(0,i)));
						}
					}
				}
			}
			// update the selection
			selFront = m_editor.docToPos(selFront);
			if( selBack.x == 0 ) {
				selBack.y--;
			}
			selBack = m_editor.docToPos(new Point(m_editor.getLineLength(selBack.y), selBack.y));
			m_editor.setCaretPosition(new Point(0, selFront.y), true);
			rc.addRedo(m_editor.setSelectionMode(SourceCodeSelection.BOUNDED));
			rc.addRedo(m_editor.setSelectionPosition(selBack));
		}
		return rc;
	}
	
	public UndoCommand indentToPrev() {
		UndoChain rc = new UndoChain("indentToPrev");
		Point p = m_editor.getCaretPosition();
		if( p.y > 0 ) {
			p.y--;
			p.x = m_editor.getSourceModel().lineFirstWord(p.y);
			p = m_editor.docToPos(p);
			while( m_editor.getCaretPosition().x < p.x ) {
				if( m_editor.getCaretPosition().x + m_editor.getTabWidth() > p.x ) {
					rc.addRedo(m_editor.insertChar(' '));
				} else {
					rc.addRedo(m_editor.insertChar('\t'));
				}
			}
		}
		return rc;
	}
	
	public UndoCommand clearSelection() {
		return m_editor.clearSelection();
	}
	
	public UndoCommand find() {
		gvid.ui.FindPanel.showDialog(gvid.ui.MainFrame.getMainFrame(), m_editor);
		return null;
	}
	
	public UndoCommand findNext() {
		String findStr = Globals.m_findSettings.m_findString;
		Point docPos = m_editor.posToDoc(m_editor.getCaretPosition());
		if( m_editor.getSelection() != null ) {
			if( isBefore( m_editor.getSelection().getSelectionStart(), m_editor.getSelection().getSelectionEnd() ) ) {
				docPos = m_editor.posToDoc(m_editor.getSelection().getSelectionEnd());
			} else {
				docPos = m_editor.posToDoc(m_editor.getSelection().getSelectionStart());
			}
		}
		Point startingPos = new Point(docPos);

		SourceCodeModel model = m_editor.getSourceModel();
		while( docPos.y <= model.getRowMaximum() ) {
			String line = model.lineToString(docPos.y).substring(docPos.x);
			int index = findSubstring(line, findStr);
			if( index != -1 ) {
				docPos.x += index;
				Point viewStart = m_editor.docToPos(docPos);
				docPos.x += findStr.length();
				Point viewEnd = m_editor.docToPos(docPos);
				m_editor.setCaretPosition( viewStart, true );
				m_editor.setSelectionMode(SourceCodeSelection.BOUNDED);
				m_editor.setSelectionPosition(viewEnd);
				return null;
			}
			docPos.y++;
			docPos.x = 0;
		}
		if( Globals.m_findSettings.m_bWrap ) {
			docPos.y = 0;
			if( startingPos.y == 0 ) {
				docPos.x = startingPos.x;
			} else {
				docPos.x = 0;
			}
			while( isBefore( docPos, startingPos ) ) {
				String line = model.lineToString(docPos.y);
				if( startingPos.y == docPos.y ) {
					line = line.substring(0, docPos.x);
				}
				int index = findSubstring(line, findStr);
				if( index != -1 ) {
					// TODO - handle whole word only in search
					if( Globals.m_findSettings.m_bWholeWord ) {
					}
					docPos.x = index;
					Point viewStart = m_editor.docToPos(docPos);
					docPos.x += findStr.length();
					Point viewEnd = m_editor.docToPos(docPos);
					m_editor.setCaretPosition( viewStart, true );
					m_editor.setSelectionMode(SourceCodeSelection.BOUNDED);
					m_editor.setSelectionPosition(viewEnd);
					return null;
				}
				docPos.y++;
				if( startingPos.y == docPos.y ) {
					docPos.x = startingPos.x;
				} else {
					docPos.x = 0;
				}
			}
		}
		return null;
	}

	public UndoCommand findPrev() {
		UndoChain rc = new UndoChain("findNext");
		return rc;
	}

	public UndoCommand findSelectionNext() {
		String str = m_editor.getSelection().getSelectedText();
		Globals.m_findSettings.m_findString = str;
		Globals.m_findSettings.addFindString(str);
		findNext();
		return null;
	}

	public UndoCommand findSelectionPrev() {
		String str = m_editor.getSelection().getSelectedText();
		Globals.m_findSettings.m_findString = str;
		Globals.m_findSettings.addFindString(str);
		findPrev();
		return null;
	}	

	// returns true if p1 is before p2
	private boolean isBefore( Point p1, Point p2 ) {
		boolean rc = false;
		if( p1.y < p2.y ) {
			rc = true;
		} else if( p1.y==p2.y && p1.x < p2.x ) {
			rc = true;
		}		
		return rc;
	}
	
	private int findSubstring( String str, String find ) {
		int rc = -1;
		for( int strIndex = 0; strIndex < str.length()-find.length()+1; strIndex++ ) {
			int findIndex = 0;
			for( ; findIndex < find.length(); findIndex++ ) {
				char strChr = str.charAt(strIndex+findIndex);
				char findChr = find.charAt(findIndex);
				if( !Globals.m_findSettings.m_bMatchCase ) {
					strChr = Character.toLowerCase(strChr);
					findChr = Character.toLowerCase(findChr);
				}
				if( strChr != findChr ) {
					break;
				}
			}
			if( findIndex == find.length() ) {
				rc = strIndex;
				break;
			}
		}
		return rc;
	}
	//{{DECLARE_CONTROLS
	//}}
}