/******************************************************************************
 * GViD Source Code Editor
 * Copyright (C) 2002  Alan Oursland
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Alan Oursland
 * alan@oursland.net
 *****************************************************************************/
package gvid.ui.splitscrollpane;

import javax.swing.*;
import javax.swing.event.*;
import java.beans.*;
import java.awt.*;

public class SplitScrollViewBase 
    extends     javax.swing.JComponent 
    implements  SplitScrollView {

    private SplitScrollModel m_model = null;
    private BoundedRangeModel m_hScrollModel = new ScrollModel(this, SwingConstants.HORIZONTAL);
    private BoundedRangeModel m_vScrollModel = new ScrollModel(this, SwingConstants.VERTICAL);
    private boolean m_bActive = false;

	public SplitScrollViewBase() {
	    setBackground(Color.white);
	    setOpaque(true);
	}

	public Component asComponent() {
		return this;
	}

	public SplitScrollView createCopy() {
	    SplitScrollView rc = new SplitScrollViewBase();
	    rc.copySettingsFrom(this);
	    return rc;
	}

    public void clearModel() {
        if( m_model != null ) {
            m_model.removeView(this);
            m_model = null;
        }
    }

    public void copySettingsFrom(SplitScrollView view) {
    	if( view != this ) {
        	m_hScrollModel.setValue(view.getHorizontalScrollModel().getValue());
        	m_vScrollModel.setValue(view.getVerticalScrollModel().getValue());
        	if( view instanceof SplitScrollViewBase ) {
        		setModel(((SplitScrollViewBase)view).m_model);
        	}
    	}
    }

    public BoundedRangeModel getHorizontalScrollModel() {
        return m_hScrollModel;
    }

    public SplitScrollModel getModel() {
        return m_model;
    }

    public BoundedRangeModel getVerticalScrollModel() {
        return m_vScrollModel;
    }

    public void setModel(SplitScrollModel model) {
        clearModel();
        if( model != null ) {
        	model.addView(this);
        	m_model = model;
        }
    }

    public void shareVerticalScrollModel( SplitScrollView view ) {
        m_vScrollModel = view.getVerticalScrollModel();
    }
    
    public void shareHorizontalScrollModel( SplitScrollView view ) {
        m_hScrollModel = view.getHorizontalScrollModel();
        
    }
    
    public int horizontalUnits( int width ) {
        return 1;
    }
    
    public int verticalUnits( int height ) {
        return 1;
    }
    
    public Rectangle getViewBounds() {
    	Rectangle rc = getBounds();
        return rc;
    }
    
    public boolean isActive() {
    	return m_bActive;
    }
    
    public void setActive(boolean active) {
    	m_bActive = active;
    }    
    
    protected int getExtent(int alignment) {
        int rc = 1;
        if( alignment == SwingConstants.HORIZONTAL ) {
            rc = horizontalUnits(getViewBounds().width)-1;
        } else {
            rc = verticalUnits(getViewBounds().height)-1;
        }
        return rc;
    }
    
    private class ScrollModel implements BoundedRangeModel {
        int m_alignment = SwingConstants.HORIZONTAL;
        int m_currentValue = 0;
        boolean m_bAdjusting = false;
        protected EventListenerList m_listenerList = new EventListenerList();
        private transient ChangeEvent changeEvent = null;
        private SplitScrollViewBase m_parent = null;
        
        public ScrollModel( SplitScrollViewBase parent, int alignment ) {
            m_alignment = alignment;
            m_parent = parent;
        }
        
        public int getMinimum() {
            int rc = 0;
            if( m_model != null ) {
                if( m_alignment == SwingConstants.HORIZONTAL ) {
                    rc = m_model.getColumnMinimum();
                } else {
                    rc = m_model.getRowMinimum();
                }
            }
            return rc;
        }
        
        public void setMinimum(int newMinimum) {
            // minimum cannot be set
        }

        public int getMaximum() {
            int rc = 1;
            if( m_model != null ) {
                if( m_alignment == SwingConstants.HORIZONTAL ) {
                    rc = m_model.getColumnMaximum();
                } else {
                    rc = m_model.getRowMaximum();
                }
            }
            return rc;
        }
        
        public void setMaximum(int newMaximum) {
            // maximum cannot be set
        }
        
        public int getValue() {
            return m_currentValue;
        }
        
        public void setValue(int newValue) {
            if( newValue < getMinimum() ) {
                newValue = getMinimum();
            } else if( newValue > getMaximum() ) {
                newValue = getMaximum();
            }
            if( m_currentValue != newValue ) {
                m_currentValue = newValue;
                fireStateChanged();
            }
        }
        
        public void setValueIsAdjusting(boolean b) {
            m_bAdjusting = b;
        }

        public boolean getValueIsAdjusting() {
            return m_bAdjusting;
        }
        
        public int getExtent() {
            return m_parent.getExtent(m_alignment);
        }
        
        public void setExtent(int newExtent) {
            // extent cannot be set
        }
        
        public void setRangeProperties(int value, int extent, int min, int max, boolean adjusting) {
            // range properties cannot be set
        }

        public void addChangeListener(ChangeListener l) {
            m_listenerList.add(ChangeListener.class, l);            
        }
        
        public void removeChangeListener(ChangeListener l) {
            m_listenerList.remove(ChangeListener.class, l);
        }
        
        public void fireStateChanged()
        {
            Object[] listeners = m_listenerList.getListenerList();
            for (int i = listeners.length - 2; i >= 0; i -= 2) {
                if (listeners[i] == ChangeListener.class) {
                    if (changeEvent == null) {
                        changeEvent = new ChangeEvent(this);
                    }
                    ((ChangeListener)listeners[i + 1]).stateChanged(changeEvent);
                }
            }
        }
    }
}